/****************************************************************************
 ****************************************************************************
 *
 * audio.c
 *
 ****************************************************************************
 ****************************************************************************/





#include "audio.h"
#include "ec64sys.h"
#include "error.h"
#include "memory.h"
#include "sync.h"



/* #define AUDIO_DEBUG_PARAMETER_LOG */
/* #define AUDIO_DEBUG_DATA_LOG */

#ifdef AUDIO_DEBUG_PARAMETER_LOG
static sd				audio_log = -1;
#endif

#ifdef AUDIO_DEBUG_DATA_LOG
static sd				audio_data_log = -1;
#endif

static d				audio_state = 0;
static sd				audio_handle;
static const b				*audio_cmdline_path = NULL;
static sw				audio_buffer[AUDIO_BUFFER_SIZE];
static d				audio_buffer_offset = 0;
static d				sid_buffer_offset = 0;
static d				audio_sample_rate;
static d				audio_samples_per_frame;
static sd				audio_output_left = 0;
static d				audio_output_size = 0;
static d				audio_time_difference = 0;

#define AUDIO_INITIALIZED		0x01
#define AUDIO_OPEN			0x02
#define AUDIO_READY			0x04
#define AUDIO_SYNC_OK			0x08



/****************************************************************************
 * audio_initialize *********************************************************
 ****************************************************************************/
void	audio_initialize(
	taudio				*audio_options)

	{
	static const b			*audio_paths[] =
					{
					STRING_TYPE AUDIO_PATH1,
					STRING_TYPE AUDIO_PATH2,
					NULL
					};
	const b				*audio_path;

	if (audio_state & AUDIO_INITIALIZED)
		error_text_exit(STRING_TYPE "!INTERNAL FAILURE! audio.c: already initialized");
	audio_state |= AUDIO_INITIALIZED;

	/* set the options, which are also needed if no audio device is found */

	audio_options->buffer = audio_buffer;
	audio_options->buffer_offset = &sid_buffer_offset;
	audio_options->buffer_size = AUDIO_BUFFER_SIZE;

	/* open the audio device */

		{
		register d		offset = 0;

		if (audio_cmdline_path != NULL) audio_handle = ec64sys_open(audio_cmdline_path, O_RDWR);
		else while (((audio_path = audio_paths[offset]) != 0) && ((audio_handle = ec64sys_open(audio_path, O_RDWR)) == -1)) offset ++;
		if (audio_handle == -1)
			{
			/*
			!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
			TODO: say the world that sound is disabled
			!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
			*/
			audio_restore();
			return;
			}
		}
	audio_state |= AUDIO_OPEN;

	/* reset the audio device and set parameters */

		{
		d			channels = 1;
		d			format = AFMT_S16_LE;
		d			blocksize;
		sd			result;

		result = ec64sys_ioctl(audio_handle, SNDCTL_DSP_RESET, NULL);
		result += ec64sys_ioctl(audio_handle, SNDCTL_DSP_SPEED, &audio_options->sample_rate);
		result += ec64sys_ioctl(audio_handle, SNDCTL_DSP_CHANNELS, &channels);
		result += ec64sys_ioctl(audio_handle, SNDCTL_DSP_SETFMT, &format);
		result -= (format != AFMT_S16_LE);
		result += ec64sys_ioctl(audio_handle, SNDCTL_DSP_GETBLKSIZE, &blocksize);
		if (result < 0)
			{
			/*
			!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
			TODO: say the world that sound is disabled
			!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
			*/
			audio_restore();
			return;
			}
		blocksize = (5 * blocksize) / 4;

		/* set the options */

		audio_sample_rate = audio_options->sample_rate;
		audio_buffer_offset = 0;
		audio_samples_per_frame = audio_sample_rate / 50;
		if ((audio_output_size = 3 * audio_samples_per_frame) < blocksize) audio_output_size = blocksize;
		}

	/* clear audio_buffer */

	memory_fill(audio_buffer, sizeof(audio_buffer), 0);

	/* now we are ready for audio output */

	audio_state |= AUDIO_READY;

#ifdef AUDIO_DEBUG_PARAMETER_LOG
	audio_log = ec64sys_open(STRING_TYPE "audio.log", (O_WRONLY | O_CREAT | O_TRUNC), (S_IRUSR | S_IWUSR));
#endif

#ifdef AUDIO_DEBUG_DATA_LOG
	audio_data_log = ec64sys_open(STRING_TYPE "audio_data.log", (O_WRONLY | O_CREAT | O_TRUNC), (S_IRUSR | S_IWUSR));
#endif
	}



/****************************************************************************
 * audio_restore ************************************************************
 ****************************************************************************/
void	audio_restore(
	void)

	{
	audio_state &= ~ AUDIO_READY;
	if (audio_state & AUDIO_OPEN)
		{
		ec64sys_close(audio_handle);
		audio_state &= ~ AUDIO_OPEN;
		}
	audio_state &= ~ AUDIO_INITIALIZED;
	}



/****************************************************************************
 * audio_set_path ***********************************************************
 ****************************************************************************/
void	audio_set_path(
	const b				*audio_path)

	{
	audio_cmdline_path = audio_path;
	}



/****************************************************************************
 * audio_play_samples *******************************************************
 ****************************************************************************/
void	audio_play_samples(
	void)

	{
	if (audio_state & AUDIO_READY)
		{
		static b		buffer_empty = 1;
		sd			bytes_left;
		register d		offset_difference;

		/* get how many samples the sounddriver has in buffer */

		if (ec64sys_ioctl(audio_handle, SNDCTL_DSP_GETODELAY, &bytes_left) == -1) return;
		if (bytes_left < 0) bytes_left = 0;
		audio_output_left = bytes_left / sizeof(sw);

		/* calculate the samples that are ready for output */

		if (sid_buffer_offset >= audio_buffer_offset) offset_difference = sid_buffer_offset - audio_buffer_offset;
		else offset_difference = (AUDIO_BUFFER_SIZE - audio_buffer_offset) + sid_buffer_offset;

		/* check if core runs too slow */

		if (offset_difference >= audio_samples_per_frame) audio_state |= AUDIO_SYNC_OK;

		/* log for debugging purposes */

#ifdef AUDIO_DEBUG_PARAMETER_LOG
		if (audio_log != -1)
			{
			sd		difference = offset_difference;

			ec64sys_write(audio_log, &difference, sizeof(difference));
			ec64sys_write(audio_log, &audio_output_left, sizeof(audio_output_left));
			ec64sys_write(audio_log, &audio_time_difference, sizeof(audio_time_difference));
			ec64sys_write(audio_log, &audio_state, sizeof(audio_state));
			}
#endif

		/* output the samples if necessary */

		if ((buffer_empty) && (offset_difference >= audio_output_size)) buffer_empty = 0;
		if ((! (buffer_empty)) && (audio_output_left < audio_output_size))
			{
			register d	length;
			register d	output_length;
			register d	offset;

			length = audio_output_size - audio_output_left;
			if (length > offset_difference)
				{
				length = offset_difference;
				buffer_empty = 1;
				}
			while (length > 0)
				{
				if ((audio_buffer_offset + length) > AUDIO_BUFFER_SIZE)
					{
					offset = audio_buffer_offset;
					output_length = AUDIO_BUFFER_SIZE - offset;
					audio_buffer_offset = 0;
					}
				else
					{
					offset = audio_buffer_offset;
					output_length = length;
					audio_buffer_offset += length;
					}
				ec64sys_write(audio_handle, &audio_buffer[offset], (output_length * sizeof(sw)));
#ifdef AUDIO_DEBUG_DATA_LOG
				ec64sys_write(audio_data_log, &audio_buffer[offset], (output_length * sizeof(sw)));
#endif
				length -= output_length;
				}

			/* calculate the time to sleep */

			if (sid_buffer_offset >= audio_buffer_offset) offset_difference = sid_buffer_offset - audio_buffer_offset;
			else offset_difference = (AUDIO_BUFFER_SIZE - audio_buffer_offset) + sid_buffer_offset;
			audio_time_difference += (SYNC_FRAME_TIME * offset_difference) / audio_samples_per_frame;

			/* check if core runs too fast (without sync) */

			if (audio_time_difference > (SYNC_FRAME_TIME + SYNC_SLEEP_TIME)) audio_time_difference = SYNC_FRAME_TIME + SYNC_SLEEP_TIME;
			}
		}
	}



/****************************************************************************
 * audio_sync ***************************************************************
 ****************************************************************************/
void	audio_sync(
	void)

	{
#ifdef AUDIO_DEBUG_PARAMETER_LOG
	if (audio_log != -1)
		{
		d		minus_one = NOT32;

		ec64sys_write(audio_log, &minus_one, sizeof(minus_one));
		ec64sys_write(audio_log, &minus_one, sizeof(minus_one));
		ec64sys_write(audio_log, &audio_time_difference, sizeof(audio_time_difference));
		ec64sys_write(audio_log, &audio_state, sizeof(audio_state));
		}
#endif
	if ((audio_state & AUDIO_READY) && (audio_state & AUDIO_SYNC_OK))
		{
		audio_time_difference = sync_delay(audio_time_difference);
		audio_state &= ~ AUDIO_SYNC_OK;
		}
	else sync_sleep();
	}
/******************************************************** ec64@unusedino.de */

