/****************************************************************************
 ****************************************************************************
 *
 * fb.c
 *
 ****************************************************************************
 ****************************************************************************/





#include "fb.h"
#include "ec64sys.h"
#include "draw.h"
#include "error.h"
#include "memory.h"



#ifdef FB_ENABLE
static d				fb_state = 0;
static sd				fb_handle;
static const b				*fb_cmdline_path = NULL;
static tscreen				fb_screeninfo_saved;
static w				fb_red_saved[256];
static w				fb_green_saved[256];
static w				fb_blue_saved[256];
static w				fb_transparent_saved[256];
static tcmap				fb_colormap_saved =
					{
					0,
					256,
					fb_red_saved,
					fb_green_saved,
					fb_blue_saved,
					fb_transparent_saved
					};
static d				fb_x_resolution;
static d				fb_y_resolution;
static d				fb_virtual_x_resolution;
static b				*fb_address;
static d				fb_mmap_size;
static d				fb_screen_lines[512];

#define FB_INITIALIZED			0x01
#define FB_OPEN				0x02
#define FB_SCREENINFO_SAVED		0x04
#define FB_COLORMAP_SAVED		0x08
#define FB_MMAP				0x10




/****************************************************************************
 * fb_initialize ************************************************************
 ****************************************************************************/
const sd	fb_initialize(
	tvidinit			*options)

	{
	static const b			*fb_paths[] =
					{
					STRING_TYPE FB_PATH1,
					STRING_TYPE FB_PATH2,
					STRING_TYPE FB_PATH3,
					STRING_TYPE FB_PATH4,
					NULL
					};
	const b				*fb_path;
	tscreen				fb_screeninfo;
	tcmap				*fb_colormap = options->colormap;

	/* open the fb device */

	if (fb_state & FB_INITIALIZED) return (-1);
	fb_state |= FB_INITIALIZED;
		{
		register d		offset = 0;

		if (fb_cmdline_path != NULL) fb_handle = ec64sys_open(fb_cmdline_path, O_RDWR);
		else while (((fb_path = fb_paths[offset]) != 0) && ((fb_handle = ec64sys_open(fb_path, O_RDWR)) == -1)) offset ++;
		if (fb_handle == -1) return (-1);
		}
	fb_state |= FB_OPEN;

	/* save current fb settings */

	if (ec64sys_ioctl(fb_handle, FBIOGET_VSCREENINFO, &fb_screeninfo_saved) == -1)
		{
		fb_restore();
		error_text_message_exit(STRING_TYPE "ioctl(FBIOGET_VSCREENINFO) failed");
		}
	if (fb_screeninfo_saved.x_offset != 0) fb_screeninfo_saved.x_offset = 0;
	if (fb_screeninfo_saved.y_offset != 0) fb_screeninfo_saved.y_offset = 0;
	fb_screeninfo = fb_screeninfo_saved;
	fb_state |= FB_SCREENINFO_SAVED;
	if (fb_screeninfo.bits_per_pixel == 4)
		{
		fb_restore();
		return (-1);
		}
	if (ec64sys_ioctl(fb_handle, FBIOGETCMAP, &fb_colormap_saved) == -1)
		{
		fb_restore();
		error_text_message_exit(STRING_TYPE "ioctl(FBIOGETCMAP) failed");
		}
	fb_state |= FB_COLORMAP_SAVED;

	/* check if fullscreen mode is enabled */

	if (options->flags == TVIDEO_FULLSCREEN)
		{
		tscreen			fb_screeninfo_fullscreen;

		memory_fill(&fb_screeninfo_fullscreen, sizeof(fb_screeninfo), 0);
		fb_screeninfo_fullscreen.x_resolution = 400;
		fb_screeninfo_fullscreen.y_resolution = 300;
		fb_screeninfo_fullscreen.virtual_x_resolution = 400;
		fb_screeninfo_fullscreen.virtual_y_resolution = 300;
		fb_screeninfo_fullscreen.bits_per_pixel = 8;
		fb_screeninfo_fullscreen.pixel_clock = 50000;
		fb_screeninfo_fullscreen.left_margin = 40;
		fb_screeninfo_fullscreen.right_margin = 20;
		fb_screeninfo_fullscreen.upper_margin = 14;
		fb_screeninfo_fullscreen.lower_margin = 1;
		fb_screeninfo_fullscreen.hsync_length = 64;
		fb_screeninfo_fullscreen.vsync_length = 2;
		fb_screeninfo_fullscreen.sync = (FB_SYNC_HOR_HIGH_ACT | FB_SYNC_VERT_HIGH_ACT);
		fb_screeninfo_fullscreen.vmode = FB_VMODE_DOUBLE;
		fb_screeninfo = fb_screeninfo_fullscreen;
		}
	else
		{
		fb_screeninfo.virtual_x_resolution = fb_screeninfo.x_resolution;
		fb_screeninfo.virtual_y_resolution = fb_screeninfo.y_resolution;
		fb_screeninfo.bits_per_pixel = 8;
		}

	/* initialize the fb device with our values */

	if (ec64sys_ioctl(fb_handle, FBIOPUT_VSCREENINFO, &fb_screeninfo) == -1)
		{
		fb_restore();
		error_text_message_exit(STRING_TYPE "ioctl(FBIOPUT_VSCREENINFO) failed");
		}
	if (fb_screeninfo.bits_per_pixel != 8)
		{
		fb_restore();
		error_text_exit(STRING_TYPE "could not change to 8bpp");
		}
	if (ec64sys_ioctl(fb_handle, FBIOPUTCMAP, fb_colormap) == -1)
		{
		fb_restore();
		error_text_message_exit(STRING_TYPE "ioctl(FBIOPUTCMAP) failed");
		}
	fb_mmap_size = fb_screeninfo.virtual_x_resolution * fb_screeninfo.virtual_y_resolution;
	fb_x_resolution = fb_screeninfo.x_resolution;
	fb_y_resolution = fb_screeninfo.y_resolution;
	fb_virtual_x_resolution = fb_screeninfo.virtual_x_resolution;

	/* mmap the fb memory */

		{
		tmmap			fb_mmap =
					{ 
					0,
					0,
					(PROT_READ | PROT_WRITE),
					MAP_SHARED,
					0,
					0
					};

		fb_mmap.length = fb_mmap_size;
		fb_mmap.handle = fb_handle;
		if ((fb_address = (b *) ec64sys_mmap(&fb_mmap)) == NULL)
			{
			fb_restore();
			error_text_message_exit(STRING_TYPE "mmap() failed");
			}
		fb_state |= FB_MMAP;
		}

	/* set the options */

	options->x_resolution = options->c64_x_resolution = fb_x_resolution;
	options->y_resolution = options->c64_y_resolution = fb_y_resolution;
	options->virtual_x_resolution = fb_virtual_x_resolution;
	options->c64_x_offset = 0;
	options->c64_y_offset = 0;
	options->c64_screen_buffer = fb_address;
	options->c64_screen_lines = fb_screen_lines;

	/* exit */

	return (0);
	}



/****************************************************************************
 * fb_restore ***************************************************************
 ****************************************************************************/
void	fb_restore(
	void)

	{
	if (fb_state & FB_SCREENINFO_SAVED) ec64sys_ioctl(fb_handle, FBIOPUT_VSCREENINFO, &fb_screeninfo_saved);
	if (fb_state & FB_COLORMAP_SAVED) ec64sys_ioctl(fb_handle, FBIOPUTCMAP, &fb_colormap_saved);
	if (fb_state & FB_MMAP)
		{
		ec64sys_munmap(fb_address, fb_mmap_size);
		fb_state &= ~ FB_MMAP;
		}
	if (fb_state & FB_OPEN)
		{
		ec64sys_close(fb_handle);
		fb_state &= ~ FB_OPEN;
		}
	fb_state &= ~ FB_INITIALIZED;
	}



/****************************************************************************
 * fb_set_path **************************************************************
 ****************************************************************************/
void	fb_set_path(
	const b				*fb_path)

	{
	fb_cmdline_path = fb_path;
	}



/****************************************************************************
 * fb_set_options ***********************************************************
 ****************************************************************************/
void	fb_set_options(
	tvidinit			*options)

	{
	options->virtual_x_resolution = fb_virtual_x_resolution;
	options->c64_screen_buffer = fb_address + (options->c64_y_offset * fb_virtual_x_resolution) + options->c64_x_offset;
	options->c64_screen_lines = fb_screen_lines;
	}



/****************************************************************************
 * fb_draw_text *************************************************************
 ****************************************************************************/
void	fb_draw_text(
	const d				x_start,
	const d				y_start,
	const b				*text)

	{
	b				*buffer;

	buffer = fb_address + (y_start * fb_virtual_x_resolution) + x_start;
	draw_text(buffer, fb_virtual_x_resolution, text);
	}



/****************************************************************************
 * fb_draw_led **************************************************************
 ****************************************************************************/
void	fb_draw_led(
	const d				x_start,
	const d				y_start,
	const b				led_state)

	{
	b				*buffer;

	buffer = fb_address + (y_start * fb_virtual_x_resolution) + x_start;
	draw_led(buffer, fb_virtual_x_resolution, led_state);
	}



/****************************************************************************
 * fb_draw_c64_screen *******************************************************
 ****************************************************************************/
void	fb_draw_c64_screen(
	void)

	{
	/* in 8bpp resolutions ec64core draws direct to the screen */
	}



/****************************************************************************
 * fb_area_clear ************************************************************
 ****************************************************************************/
void	fb_area_clear(
	const d				x_start,
	const d				y_start,
	const d				x_end,
	const d				y_end)

	{
	if (x_start < x_end)
		{
		d			y_position;
		d			x_length = x_end - x_start;
		b			*buffer = fb_address + (y_start * fb_virtual_x_resolution) + x_start;

		for (y_position = y_start; y_position < y_end; y_position ++)
			{
			memory_fill(buffer, x_length, 0);
			buffer += fb_virtual_x_resolution;
			}
		}
	}



/****************************************************************************
 * fb_area_move *************************************************************
 ****************************************************************************/
void	fb_area_move(
	const d				x_start,
	const d				y_start,
	const d				x_end,
	const d				y_end,
	const d				x_destination,
	const d				y_destination)

	{
	if (x_start != x_destination)
		error_text_exit(STRING_TYPE "!INTERNAL FAILURE! fb.c: fb_area_move() failed");
	if (x_start < x_end)
		{
		d			y_position;
		d			x_length = x_end - x_start;

		if (y_start < y_destination)
			{
			b		*source_buffer = fb_address + ((y_end - 1) * fb_virtual_x_resolution) + x_start;
			b		*destination_buffer = fb_address + ((y_destination + y_end - y_start - 1) * fb_virtual_x_resolution) + x_destination;

			for (y_position = y_end - 1; y_position >= y_start; y_position --)
				{
				memory_copy(destination_buffer, source_buffer, x_length);
				source_buffer -= fb_virtual_x_resolution;
				destination_buffer -= fb_virtual_x_resolution;
				}
			}
		else
			{
			b		*source_buffer = fb_address + (y_start * fb_virtual_x_resolution) + x_start;
			b		*destination_buffer = fb_address + (y_destination * fb_virtual_x_resolution) + x_destination;

			for (y_position = y_start; y_position < y_end; y_position ++)
				{
				memory_copy(destination_buffer, source_buffer, x_length);
				source_buffer += fb_virtual_x_resolution;
				destination_buffer += fb_virtual_x_resolution;
				}
			}
		}
	}



#ifdef FB_ENABLE_SCREENSHOT_SUPPORT
/****************************************************************************
 * fb_get_line **************************************************************
 ****************************************************************************/
void	fb_get_line(
	b				*buffer,
	const d				line,
	const d				length)

	{
	if (line > fb_y_resolution) return;
	memory_copy(buffer, (fb_address + fb_virtual_x_resolution * line), length);
	}
#endif /* FB_ENABLE_SCREENSHOT_SUPPORT */
#endif /* FB_ENABLE */
/******************************************************** ec64@unusedino.de */

