/****************************************************************************
 ****************************************************************************
 *
 * tty.c
 *
 ****************************************************************************
 ****************************************************************************/





#include "tty.h"
#include "ec64sys.h"
#include "ec64io.h"
#include "string.h"
#include "memory.h"
#include "error.h"



static d				tty_state = 0;
static sd				tty_handle = STDIN;
static d				tty_fl_saved;
static ttermios				tty_termios_saved;
static tvtmode				tty_vtmode_saved;
static d				tty_kbmode_saved;
static w				tty_kbentries_saved[1 << (KG_ALT + 1)][NR_KEYS];
static d				tty_kdmode_saved;

#define TTY_INITIALIZED		0x0001
#define TTY_OPEN		0x0002
#define TTY_FL_SAVED		0x0004
#define TTY_TERMIOS_SAVED	0x0008
#define TTY_VTMODE_SAVED	0x0010
#define TTY_KBMODE_SAVED	0x0020
#define TTY_KBENTRIES_SAVED	0x0040
#define TTY_LOCKED		0x0080
#define TTY_CLEAR_SCREEN	0x0100
#define TTY_CURSOR_OFF		0x0200
#define TTY_BLANKING_OFF	0x0400
#define TTY_KDMODE_SAVED	0x0800



/****************************************************************************
 * tty_cursor_on ************************************************************
 ****************************************************************************/
void	tty_cursor_on(
	void)

	{
	if (tty_state & TTY_CURSOR_OFF)
		{
		ec64sys_write(STDOUT, "\033[?25h", 6);
		tty_state &= ~ TTY_CURSOR_OFF;
		}
	}



/****************************************************************************
 * tty_cursor_off ***********************************************************
 ****************************************************************************/
void	tty_cursor_off(
	void)

	{
	if (! (tty_state & TTY_CURSOR_OFF))
		{
		ec64sys_write(STDOUT, "\033[?25l", 6);
		tty_state |= TTY_CURSOR_OFF;
		}
	}



/****************************************************************************
 * tty_blanking_on **********************************************************
 ****************************************************************************/
void	tty_blanking_on(
	void)

	{
	if (tty_state & TTY_BLANKING_OFF)
		{
		ec64sys_write(STDOUT, "\033[9;5]", 6);
		tty_state &= ~ TTY_BLANKING_OFF;
		}
	}



/****************************************************************************
 * tty_blanking_off *********************************************************
 ****************************************************************************/
void	tty_blanking_off(
	void)

	{
	if (! (tty_state & TTY_BLANKING_OFF))
		{
		ec64sys_write(STDOUT, "\033[9;0]", 6);
		tty_state |= TTY_BLANKING_OFF;
		}
	}



/****************************************************************************
 * tty_clear_screen *********************************************************
 ****************************************************************************/
void	tty_clear_screen(
	void)

	{
	ec64sys_write(STDOUT, "\033[1;1H\033[2J", 10);
	tty_state |= TTY_CLEAR_SCREEN;
	}



/****************************************************************************
 * tty_initialize ***********************************************************
 ****************************************************************************/
void	tty_initialize(
	void)

	{
	ttermios			tty_termios;
	tvtmode				tty_vtmode;

	/* already initialized ? */

	if (tty_state & TTY_INITIALIZED)
		error_text_exit(STRING_TYPE "!INTERNAL FAILURE! tty.c: already initialized");
	tty_state |= TTY_INITIALIZED;

	/* try to open /dev/tty */

	if ((tty_handle = ec64sys_open(STRING_TYPE "/dev/tty", O_RDWR)) == -1) tty_handle = STDIN;
	else tty_state |= TTY_OPEN;

	/* check if we are on local console */

		{
		d			number = 12;
        
		if ((ec64sys_ioctl(STDIN, TIOCLINUX, &number) == -1) ||
		    (ec64sys_ioctl(STDOUT, TIOCLINUX, &number) == -1) ||
		    (ec64sys_ioctl(tty_handle, TIOCLINUX, &number) == -1))
			error_text_exit(STRING_TYPE "not on local console");
		}
#ifndef EC64_ENABLE_KDSETMODE

	/* cursor off */

	tty_cursor_off();
	tty_blanking_off();
#endif
	tty_clear_screen();

	/* save current flags */

	if ((tty_fl_saved = ec64sys_fcntl(STDIN, F_GETFL)) == -1)
		error_text_message_exit(STRING_TYPE "fcntl(F_GETFL) failed");
	tty_state |= TTY_FL_SAVED;

	/* make reads from STDIN nonblocking */

	if (ec64sys_fcntl(STDIN, F_SETFL, (tty_fl_saved | O_NONBLOCK)) == -1)
		error_text_message_exit(STRING_TYPE "fcntl(F_SETFL) failed");

	/* save terminal state */

	if (ec64sys_ioctl(tty_handle, TCGETS, &tty_termios_saved) == -1)
		error_text_message_exit(STRING_TYPE "ioctl(TCGETS) failed");
	tty_termios = tty_termios_saved;
	tty_state |= TTY_TERMIOS_SAVED;
	if (ec64sys_ioctl(tty_handle, VT_GETMODE, &tty_vtmode_saved) == -1)
		error_text_message_exit(STRING_TYPE "ioctl(VT_GETMODE) failed");
	tty_vtmode = tty_vtmode_saved;
	tty_state |= TTY_VTMODE_SAVED;
	if (ec64sys_ioctl(tty_handle, KDGKBMODE, &tty_kbmode_saved) == -1)
		error_text_message_exit(STRING_TYPE "ioctl(KDGKBMODE) failed");
	tty_state |= TTY_KBMODE_SAVED;

		{
		d			table;
		d			index;
		tkbentry		kbentry;

		for (table = 0; table < (1 << (KG_ALT + 1)); table ++)
			{
			kbentry.table = table;
			for (index = 0; index < NR_KEYS; index ++)
				{
				kbentry.index = index;
				if (ec64sys_ioctl(tty_handle, KDGKBENT, &kbentry) == -1)
					error_text_message_exit(STRING_TYPE "ioctl(KDGKBENT) failed");
				tty_kbentries_saved[table][index] = kbentry.value;
				}
			}
		}
	tty_state |= TTY_KBENTRIES_SAVED;

#ifdef EC64_ENABLE_KDSETMODE
	/* save the current mode */

	if (ec64sys_ioctl(tty_handle, KDGETMODE, &tty_kdmode_saved) == -1)
		error_text_message_exit(STRING_TYPE "ioctl(KDGETMODE) failed");
	tty_state |= TTY_KDMODE_SAVED;
#endif

	/* set terminal values */

	tty_termios.input_flags &= ~ (ISTRIP | INLCR | ICRNL | IGNCR | IXON | IXOFF);
	tty_termios.local_flags &= ~ (ECHO | ICANON | ISIG);
	if (ec64sys_ioctl(tty_handle, TCSETSW, &tty_termios) == -1)
		error_text_message_exit(STRING_TYPE "ioctl(TCSETSW) failed");
	tty_state |= TTY_LOCKED;
#ifndef EC64_ENABLE_KDSETMODE
	tty_vtmode.mode = VT_PROCESS;
	if (ec64sys_ioctl(tty_handle, VT_SETMODE, &tty_vtmode) == -1)
		error_text_message_exit(STRING_TYPE "ioctl(VT_SETMODE) failed");
#endif
	if (ec64sys_ioctl(tty_handle, KDSKBMODE, K_RAW) == -1)
		error_text_message_exit(STRING_TYPE "ioctl(KDSKBMODE) failed");
#ifdef EC64_ENABLE_KDSETMODE
	if (ec64sys_ioctl(tty_handle, KDSETMODE, KD_GRAPHICS) == -1)
		error_text_message_exit(STRING_TYPE "ioctl(KDSETMODE) failed");
#endif
	}



/****************************************************************************
 * tty_restore **************************************************************
 ****************************************************************************/
void	tty_restore(
	void)

	{
#ifdef EC64_ENABLE_KDSETMODE
	if (tty_state & TTY_KDMODE_SAVED) ec64sys_ioctl(tty_handle, KDSETMODE, tty_kdmode_saved);
#endif
	if (tty_state & TTY_TERMIOS_SAVED) ec64sys_ioctl(tty_handle, TCSETSW, &tty_termios_saved);
	if (tty_state & TTY_VTMODE_SAVED) ec64sys_ioctl(tty_handle, VT_SETMODE, &tty_vtmode_saved);
	if (tty_state & TTY_KBMODE_SAVED) ec64sys_ioctl(tty_handle, KDSKBMODE, tty_kbmode_saved);
	if (tty_state & TTY_FL_SAVED) ec64sys_fcntl(STDIN, F_SETFL, tty_fl_saved);
	tty_state &= ~ TTY_LOCKED;
#ifndef EC64_ENABLE_KDSETMODE
	tty_cursor_on();
	tty_blanking_on();
#endif
	if (tty_state & TTY_OPEN) ec64sys_close(tty_handle);
	if (tty_state & TTY_CLEAR_SCREEN) tty_clear_screen();
	tty_state &= ~ TTY_INITIALIZED;
	}



/****************************************************************************
 * tty_get_character ********************************************************
 ****************************************************************************/
#define WRITE_KEYCODE(x) \
	{ \
	queue[queue_write_offset ++] = x; \
	if (queue_write_offset >= TTY_KEYCODE_QUEUE_LENGTH) queue_write_offset = 0; \
	}
#define READ_KEYCODE(x) \
	{ \
	x = queue[queue_read_offset ++]; \
	if (queue_read_offset >= TTY_KEYCODE_QUEUE_LENGTH) queue_read_offset = 0; \
	}
	

d	tty_get_character(
	void)

	{
	static d			metakeys = 0;
	static b			mode = 0;
	static b			last_keycode = 0;
	static b			queue_read_offset = 0;
	static b			queue_write_offset = 0;
	static b			queue[TTY_KEYCODE_QUEUE_LENGTH];
	b				scancode;
	b				keycode = last_keycode;

	while (ec64sys_read(STDIN, &scancode, 1) == 1)
		{
		if (mode == 1)
			{
			switch (scancode)
				{
				case 0x1d:
					metakeys |= ((1 << KG_CTRL) | (1 << KG_CTRLR));
					break;
				case 0x9d:
					metakeys &= ~ (1 << KG_CTRLR);
					if (! (metakeys & (1 << KG_CTRLL))) metakeys &= ~ (1 << KG_CTRL);
					break;
				case 0x38:
					metakeys |= (1 << KG_ALTGR);
					break;
				case 0xb8:
					metakeys &= ~ (1 << KG_ALTGR);
					break;
				case 0x48:
					WRITE_KEYCODE(0x60)
					break;
				case 0xc8:
					WRITE_KEYCODE(0xe0)
					break;
				case 0x50:
					WRITE_KEYCODE(0x61)
					break;
				case 0xd0:
					WRITE_KEYCODE(0xe1)
					break;
				case 0x4b:
					WRITE_KEYCODE(0x62)
					break;
				case 0xcb:
					WRITE_KEYCODE(0xe2)
					break;
				case 0x4d:
					WRITE_KEYCODE(0x63)
					break;
				case 0xcd:
					WRITE_KEYCODE(0xe3)
					break;
				case 0x47:
					WRITE_KEYCODE(0x64)
					break;
				case 0xc7:
					WRITE_KEYCODE(0xe4)
					break;
				case 0x52:
					WRITE_KEYCODE(0x65)
					break;
				case 0xd2:
					WRITE_KEYCODE(0xe5)
					break;
				case 0x53:
					WRITE_KEYCODE(0x66)
					break;
				case 0xd3:
					WRITE_KEYCODE(0xe6)
					break;
				case 0x49:
					WRITE_KEYCODE(0x67)
					break;
				case 0xc9:
					WRITE_KEYCODE(0xe7)
					break;
				}
			mode = 0;
			continue;
			}
		if (mode == 2)
			{
			mode ++;
			continue;
			}
		if (mode == 3)
			{
			switch (scancode)
				{
				case 0x45:
					WRITE_KEYCODE(0x7f)
					break;
				case 0xc5:
					WRITE_KEYCODE(0xff)
					break;
				}
			mode = 0;
			continue;
			}
		switch (scancode)
			{
			case 0xe0:
				mode = 1;
				break;
			case 0xe1:
				mode = 2;
				break;
			case 0x1d:
				metakeys |= ((1 << KG_CTRL) | (1 << KG_CTRLL));
				break;
			case 0x9d:
				metakeys &= ~ (1 << KG_CTRLL);
				if (! (metakeys & (1 << KG_CTRLR))) metakeys &= ~ (1 << KG_CTRL);
				break;
			case 0x2a:
				metakeys |= ((1 << KG_SHIFT) | (1 << KG_SHIFTL));
				break;
			case 0xaa:
				metakeys &= ~ (1 << KG_SHIFTL);
				if (! (metakeys & (1 << KG_SHIFTR))) metakeys &= ~ (1 << KG_SHIFT);
				break;
			case 0x36:
				metakeys |= ((1 << KG_SHIFT) | (1 << KG_SHIFTR));
				break;
			case 0xb6:
				metakeys &= ~ (1 << KG_SHIFTR);
				if (! (metakeys & ((1 << KG_SHIFTR) | (1 << KG_SHIFTR)))) metakeys &= ~ (1 << KG_SHIFT);
				break;
			case 0x38:
				metakeys |= (1 << KG_ALT);
				break;
			case 0xb8:
				metakeys &= ~ (1 << KG_ALT);
				break;
			case 0x01:
				if (metakeys & (1 << KG_CTRL)) ec64io_end();
				break;
			default:
				WRITE_KEYCODE(scancode)
				break;
			}
		}
	if (queue_write_offset != queue_read_offset)
		{
		READ_KEYCODE(keycode)
		last_keycode = keycode;
		}
	return ((metakeys << 8) | keycode);
	}



/****************************************************************************
 * tty_translate_character **************************************************
 ****************************************************************************/
b	tty_translate_character(
	const d				character)

	{
	d				table = (character >> 8) & ((1 << (KG_ALT + 1)) - 1);
	d				index = character & 0xff;

	if (index & 0x80) return (0);
	if ((table == 0) && (index == 0x1c)) return (0x0a);
	return (tty_kbentries_saved[table][index]);
	}



/****************************************************************************
 * tty_print_text ***********************************************************
 ****************************************************************************/
void	tty_print_text(
	const b				*text)

	{
	if (! (tty_state & TTY_LOCKED)) ec64sys_write(STDOUT, text, string_length(text));
	}



/****************************************************************************
 * tty_print_error **********************************************************
 ****************************************************************************/
void	tty_print_error(
	const b				*error)

	{
	if (! (tty_state & TTY_LOCKED)) ec64sys_write(STDERR, error, string_length(error));
	}
/******************************************************** ec64@unusedino.de */

