/****************************************************************************
 ****************************************************************************
 *
 * vga.c
 *
 ****************************************************************************
 ****************************************************************************/





#include "vga.h"
#include "ec64sys.h"
#include "draw.h"
#include "error.h"
#include "memory.h"
#include "string.h"





#ifdef VGA_ENABLE
static d				vga_state = 0;
static d				vga_crt_index = VGA_CRT_INDEX;
static d				vga_crt_data = VGA_CRT_DATA;
static d				vga_input_status = VGA_INPUT_STATUS;
static b				vga_crt_registers_saved[VGA_CRT_REGISTERS];
static b				vga_attribute_registers_saved[VGA_ATTRIBUTE_REGISTERS];
static b				vga_graphic_registers_saved[VGA_GRAPHIC_REGISTERS];
static b				vga_sequencer_registers_saved[VGA_SEQUENCER_REGISTERS];
static b				vga_misc_registers_saved[VGA_MISC_REGISTERS];
static w				vga_red_saved[256];
static w				vga_green_saved[256];
static w				vga_blue_saved[256];
static tcmap				vga_colormap_saved =
					{
					0,
					256,
					vga_red_saved,
					vga_green_saved,
					vga_blue_saved,
					vga_blue_saved
					};
static b				vga_charset_saved[VGA_CHARSET_SIZE];
static const b				*vga_cmdline_mode_string = NULL;
static d				vga_mode = 0;
static d				vga_x_resolution;
static d				vga_y_resolution;
static d				vga_virtual_x_resolution;
static b				*vga_address;
static d				vga_mmap_size;
static d				vga_screen_lines[512];
static b				vga_screen_buffer[(VGA_X_RESOLUTION * VGA_Y_RESOLUTION)];
static d				vga_c64_x_offset;
static d				vga_c64_y_offset;
static b				vga_draw_buffer[(DRAW_X_CHARACTER_SIZE * VGA_X_RESOLUTION)];
static struct
	{
	const b				*mode_string;
	d				x_resolution;
	d				y_resolution;
	b				crt_registers[VGA_CRT_REGISTERS];
	b				attribute_registers[VGA_ATTRIBUTE_REGISTERS];
	b				graphic_registers[VGA_GRAPHIC_REGISTERS];
	b				sequencer_registers[VGA_SEQUENCER_REGISTERS];
	b				misc_registers[VGA_MISC_REGISTERS];
	}				vga_modes[] =
	{

		/* 360x240 (default mode) */

		{
			STRING_TYPE "360x240",
			360, 240,
			{ 0x6b, 0x59, 0x5a, 0x8e, 0x5e, 0x8a, 0x0d, 0x3e,
			  0x00, 0x41, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
			  0xea, 0xac, 0xdf, 0x2d, 0x00, 0xe7, 0x06, 0xe3 },
			{ 0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07,
			  0x08, 0x09, 0x0a, 0x0b, 0x0c, 0x0d, 0x0e, 0x0f,
			  0x41, 0x00, 0x0f, 0x00, 0x00 },
			{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x40, 0x05, 0x0f,
			  0xff },
			{ 0x03, 0x01, 0x0f, 0x00, 0x06 },
			{ 0xe7 },
		},

		/* 320x240 */

		{
			STRING_TYPE "320x240",
			320, 240,
			{ 0x5f, 0x4f, 0x50, 0x82, 0x54, 0x80, 0x0d, 0x3e,
                          0x00, 0x41, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
                          0xea, 0xac, 0xdf, 0x28, 0x00, 0xe7, 0x06, 0xe3 },
			{ 0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07,
                          0x08, 0x09, 0x0a, 0x0b, 0x0c, 0x0d, 0x0e, 0x0f,
                          0x41, 0x00, 0x0f, 0x00, 0x00 },
			{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x40, 0x05, 0x0f,
			  0xff },
			{ 0x03, 0x01, 0x0f, 0x00, 0x06 },
			{ 0xe3 },
		},

		/* end */

		{ NULL }
	};

#define VGA_INITIALIZED			0x00000001
#define VGA_IOPERM			0x00000002
#define VGA_IOPERM_MONOCHROME1		0x00000004
#define VGA_IOPERM_MONOCHROME2		0x00000008
#define VGA_MMAP			0x00000010
#define VGA_CRT_REGISTERS_SAVED		0x00000100
#define VGA_ATTRIBUTE_REGISTERS_SAVED	0x00000200
#define VGA_GRAPHIC_REGISTERS_SAVED	0x00000400
#define VGA_SEQUENCER_REGISTERS_SAVED	0x00000800
#define VGA_MISC_REGISTERS_SAVED	0x00001000
#define VGA_COLORMAP_SAVED		0x00002000
#define VGA_CHARSET_SAVED		0x00004000



/****************************************************************************
 * vga_get_crt_registers ****************************************************
 ****************************************************************************/
void	vga_get_crt_registers(
	b				*crt_registers)

	{
	d				offset;

	for (offset = 0; offset < VGA_CRT_REGISTERS; offset ++)
		{
		ec64sys_outb(vga_crt_index, offset);
		crt_registers[offset] = ec64sys_inb(vga_crt_data);
		}
	}



/****************************************************************************
 * vga_get_attribute_registers **********************************************
 ****************************************************************************/
void	vga_get_attribute_registers(
	b				*attribute_registers)

	{
	d				offset;

	for (offset = 0; offset < VGA_ATTRIBUTE_REGISTERS; offset ++)
		{
		ec64sys_inb(vga_input_status);
		ec64sys_outb(VGA_ATTRIBUTE_INDEX, offset);
		attribute_registers[offset] = ec64sys_inb(VGA_ATTRIBUTE_DATA_READ);
		}
	}



/****************************************************************************
 * vga_get_graphic_registers ************************************************
 ****************************************************************************/
void	vga_get_graphic_registers(
	b				*graphic_registers)

	{
	d				offset;

	for (offset = 0; offset < VGA_GRAPHIC_REGISTERS; offset ++)
		{
		ec64sys_outb(VGA_GRAPHIC_INDEX, offset);
		graphic_registers[offset] = ec64sys_inb(VGA_GRAPHIC_DATA);
		}
	}



/****************************************************************************
 * vga_get_sequencer_registers **********************************************
 ****************************************************************************/
void	vga_get_sequencer_registers(
	b				*sequencer_registers)

	{
	d				offset;

	for (offset = 0; offset < VGA_SEQUENCER_REGISTERS; offset ++)
		{
		ec64sys_outb(VGA_SEQUENCER_INDEX, offset);
		sequencer_registers[offset] = ec64sys_inb(VGA_SEQUENCER_DATA);
		}
	}



/****************************************************************************
 * vga_get_misc_registers ***************************************************
 ****************************************************************************/
void	vga_get_misc_registers(
	b				*misc_registers)

	{
	misc_registers[0] = ec64sys_inb(VGA_MISC_DATA_READ);
	}



/****************************************************************************
 * vga_get_colormap *********************************************************
 ****************************************************************************/
void	vga_get_colormap(
	const tcmap			*colormap)

	{
	d				offset;
	d				length = colormap->length;

	ec64sys_outb(VGA_PEL_INDEX_READ, colormap->offset);
	if (length + colormap->offset > 256) length = 256 - colormap->offset;
	for (offset = 0; offset < length; offset ++)
		{
		colormap->red[offset] = (w) (ec64sys_inb(VGA_PEL_DATA)) << 10;
		colormap->green[offset] = (w) (ec64sys_inb(VGA_PEL_DATA)) << 10;
		colormap->blue[offset] = (w) (ec64sys_inb(VGA_PEL_DATA)) << 10;
		}
	}



/****************************************************************************
 * vga_save_registers *******************************************************
 ****************************************************************************/
void	vga_save_registers(
	void)

	{
	if (! (vga_state & VGA_CRT_REGISTERS_SAVED))
		{
		vga_get_crt_registers(vga_crt_registers_saved);
		vga_state |= VGA_CRT_REGISTERS_SAVED;
		}
	if (! (vga_state & VGA_ATTRIBUTE_REGISTERS_SAVED))
		{
		vga_get_attribute_registers(vga_attribute_registers_saved);
		vga_state |= VGA_ATTRIBUTE_REGISTERS_SAVED;
		}
	if (! (vga_state & VGA_GRAPHIC_REGISTERS_SAVED))
		{
		vga_get_graphic_registers(vga_graphic_registers_saved);
		vga_state |= VGA_GRAPHIC_REGISTERS_SAVED;
		}
	if (! (vga_state & VGA_SEQUENCER_REGISTERS_SAVED))
		{
		vga_get_sequencer_registers(vga_sequencer_registers_saved);
		vga_state |= VGA_SEQUENCER_REGISTERS_SAVED;
		}
	if (! (vga_state & VGA_MISC_REGISTERS_SAVED))
		{
		vga_get_misc_registers(vga_misc_registers_saved);
		vga_state |= VGA_MISC_REGISTERS_SAVED;
		}
	if (! (vga_state & VGA_COLORMAP_SAVED))
		{
		vga_get_colormap(&vga_colormap_saved);
		vga_state |= VGA_COLORMAP_SAVED;
		}
	}



/****************************************************************************
 * vga_set_crt_registers ****************************************************
 ****************************************************************************/
void	vga_set_crt_registers(
	b				*crt_registers)

	{
	d				offset;

	/* deprotect CRT Registers 0-7 */

	ec64sys_outb(vga_crt_index, 0x11);
	ec64sys_outb(vga_crt_data, ec64sys_inb(vga_crt_data) & 0x7f);

	/* write to the registers */

	for (offset = 0; offset < VGA_CRT_REGISTERS; offset ++)
		{
		ec64sys_outb(vga_crt_index, offset);
		ec64sys_outb(vga_crt_data, crt_registers[offset]);
		}

	/* protect CRT Registers 0-7 */

	ec64sys_outb(vga_crt_index, 0x11);
	ec64sys_outb(vga_crt_data, ec64sys_inb(VGA_CRT_DATA) | 0x80);
	}



/****************************************************************************
 * vga_set_attribute_registers **********************************************
 ****************************************************************************/
void	vga_set_attribute_registers(
	b				*attribute_registers)

	{
	d				offset;

	for (offset = 0; offset < VGA_ATTRIBUTE_REGISTERS; offset ++)
		{
		ec64sys_inb(vga_input_status);
		ec64sys_outb(VGA_ATTRIBUTE_INDEX, offset);
		ec64sys_outb(VGA_ATTRIBUTE_DATA_WRITE, attribute_registers[offset]);
		}
	}



/****************************************************************************
 * vga_set_graphic_registers ************************************************
 ****************************************************************************/
void	vga_set_graphic_registers(
	b				*graphic_registers)

	{
	d				offset;

	for (offset = 0; offset < VGA_GRAPHIC_REGISTERS; offset ++)
		{
		ec64sys_outb(VGA_GRAPHIC_INDEX, offset);
		ec64sys_outb(VGA_GRAPHIC_DATA, graphic_registers[offset]);
		}
	}



/****************************************************************************
 * vga_set_sequencer_registers **********************************************
 ****************************************************************************/
void	vga_set_sequencer_registers(
	b				*sequencer_registers)

	{
	d				offset;

	/* synchronous reset on */

	ec64sys_outb(VGA_SEQUENCER_INDEX, 0);
	ec64sys_outb(VGA_SEQUENCER_DATA, 1);

	/* write to the registers */

	ec64sys_outb(VGA_SEQUENCER_INDEX, 1);
	ec64sys_outb(VGA_SEQUENCER_DATA, sequencer_registers[1] | 0x20);

	for (offset = 2; offset < VGA_SEQUENCER_REGISTERS; offset ++)
		{
		ec64sys_outb(VGA_SEQUENCER_INDEX, offset);
		ec64sys_outb(VGA_SEQUENCER_DATA, sequencer_registers[offset]);
		}

	/* synchronous reset off */

	ec64sys_outb(VGA_SEQUENCER_INDEX, 0);
	ec64sys_outb(VGA_SEQUENCER_DATA, 3);
	}



/****************************************************************************
 * vga_set_misc_registers ***************************************************
 ****************************************************************************/
void	vga_set_misc_registers(
	b				*misc_registers)

	{
	ec64sys_outb(VGA_MISC_DATA_WRITE, misc_registers[0]);
	}



/****************************************************************************
 * vga_set_colormap *********************************************************
 ****************************************************************************/
void	vga_set_colormap(
	const tcmap			*colormap)

	{
	d				offset;
	d				length = colormap->length;

	ec64sys_outb(VGA_PEL_INDEX_WRITE, colormap->offset);
	if (length + colormap->offset > 256) length = 256 - colormap->offset;
	for (offset = 0; offset < length; offset ++)
		{
		ec64sys_outb(VGA_PEL_DATA, (colormap->red[offset] >> 10));
		ec64sys_outb(VGA_PEL_DATA, (colormap->green[offset] >> 10));
		ec64sys_outb(VGA_PEL_DATA, (colormap->blue[offset] >> 10));
		}
	}



/****************************************************************************
 * vga_screen_on ************************************************************
 ****************************************************************************/
void	vga_screen_on(
	void)

	{

	/* turn on the screen */

	ec64sys_outb(VGA_SEQUENCER_INDEX, 1);
	ec64sys_outb(VGA_SEQUENCER_DATA, ec64sys_inb(VGA_SEQUENCER_DATA) & 0xdf);

	/* enable video output */

	ec64sys_inb(vga_input_status);
	ec64sys_outb(VGA_ATTRIBUTE_DATA_WRITE, 0x20);
	}


/****************************************************************************
 * vga_switch_mode **********************************************************
 ****************************************************************************/
void	vga_switch_mode(
	void)

	{
	vga_set_misc_registers(vga_modes[vga_mode].misc_registers);
	vga_set_sequencer_registers(vga_modes[vga_mode].sequencer_registers);
	vga_set_crt_registers(vga_modes[vga_mode].crt_registers);
	vga_set_graphic_registers(vga_modes[vga_mode].graphic_registers);
	vga_set_attribute_registers(vga_modes[vga_mode].attribute_registers);
	vga_screen_on();
	}


/****************************************************************************
 * vga_select_read_plane ****************************************************
 ****************************************************************************/
void	vga_select_read_plane(
	const b				read_plane,
	const b				read_mode)

	{

	/* select read mode */

	ec64sys_outb(VGA_GRAPHIC_INDEX, 5);
	ec64sys_outb(VGA_GRAPHIC_DATA, ((ec64sys_inb(VGA_GRAPHIC_DATA) & 0xf7) | ((read_mode << 3) & 0x08)));

	/* select read plane */

	ec64sys_outb(VGA_GRAPHIC_INDEX, 4);
	ec64sys_outb(VGA_GRAPHIC_DATA, (read_plane & 0x03));
	}



/****************************************************************************
 * vga_select_write_plane ***************************************************
 ****************************************************************************/
void	vga_select_write_plane(
	const b				write_plane,
	const b				write_mode)

	{

	/* enable set/reset = 0 */

	ec64sys_outb(VGA_GRAPHIC_INDEX, 1);
	ec64sys_outb(VGA_GRAPHIC_DATA, 0);

	/* logical operation = none, rotate = 0 */

	ec64sys_outb(VGA_GRAPHIC_INDEX, 3);
	ec64sys_outb(VGA_GRAPHIC_DATA, 0);

	/* select write mode */

	ec64sys_outb(VGA_GRAPHIC_INDEX, 5);
	ec64sys_outb(VGA_GRAPHIC_DATA, ((ec64sys_inb(VGA_GRAPHIC_DATA) & 0xfc) | (write_mode & 0x03)));

	/* bitmask = 0xff */

	ec64sys_outb(VGA_GRAPHIC_INDEX, 8);
	ec64sys_outb(VGA_GRAPHIC_DATA, 0xff);

	/* select write_plane */

	/* NOTE: you can select multiple planes to be written at once, so
	   this selection method uses a bit for every plane (in contrast to
	   the selection of the read plane, which wants the plane number) */

	ec64sys_outb(VGA_SEQUENCER_INDEX, 2);
	ec64sys_outb(VGA_SEQUENCER_DATA, (write_plane & 0x0f));
	}



/****************************************************************************
 * vga_restore_registers ****************************************************
 ****************************************************************************/
void	vga_restore_registers(
	void)

	{
	if (vga_state & VGA_MISC_REGISTERS_SAVED)
		vga_set_misc_registers(vga_misc_registers_saved);
	if (vga_state & VGA_SEQUENCER_REGISTERS_SAVED)
		vga_set_sequencer_registers(vga_sequencer_registers_saved);
	if (vga_state & VGA_CRT_REGISTERS_SAVED)
		vga_set_crt_registers(vga_crt_registers_saved);
	if (vga_state & VGA_GRAPHIC_REGISTERS_SAVED)
		vga_set_graphic_registers(vga_graphic_registers_saved);
	if (vga_state & VGA_ATTRIBUTE_REGISTERS_SAVED)
		vga_set_attribute_registers(vga_attribute_registers_saved);
	vga_screen_on();
	if (vga_state & VGA_COLORMAP_SAVED)
		vga_set_colormap(&vga_colormap_saved);
	}



/****************************************************************************
 * vga_save_charset *********************************************************
 ****************************************************************************/
void	vga_save_charset(
	void)

	{
	if (! (vga_state & VGA_CHARSET_SAVED))
		{
		vga_select_read_plane(0, 0);
		memory_copy(vga_charset_saved, vga_address, (VGA_CHARSET_SIZE / VGA_PLANES));
		vga_select_read_plane(1, 0);
		memory_copy((vga_charset_saved + (VGA_CHARSET_SIZE / VGA_PLANES)), vga_address, (VGA_CHARSET_SIZE / VGA_PLANES));
		vga_select_read_plane(2, 0);
		memory_copy((vga_charset_saved + 2 * (VGA_CHARSET_SIZE / VGA_PLANES)), vga_address, (VGA_CHARSET_SIZE / VGA_PLANES));
		vga_select_read_plane(3, 0);
		memory_copy((vga_charset_saved + 3 * (VGA_CHARSET_SIZE / VGA_PLANES)), vga_address, (VGA_CHARSET_SIZE / VGA_PLANES));
		vga_state |= VGA_CHARSET_SAVED;
		}
	}



/****************************************************************************
 * vga_restore_charset ******************************************************
 ****************************************************************************/
void	vga_restore_charset(
	void)

	{
	if (vga_state & VGA_CHARSET_SAVED)
		{
		vga_select_write_plane(1, 0);
		memory_copy(vga_address, vga_charset_saved, (VGA_CHARSET_SIZE / VGA_PLANES));
		vga_select_write_plane(2, 0);
		memory_copy(vga_address, (vga_charset_saved + (VGA_CHARSET_SIZE / VGA_PLANES)), (VGA_CHARSET_SIZE / VGA_PLANES));
		vga_select_write_plane(4, 0);
		memory_copy(vga_address, (vga_charset_saved + 2 * (VGA_CHARSET_SIZE / VGA_PLANES)), (VGA_CHARSET_SIZE / VGA_PLANES));
		vga_select_write_plane(8, 0);
		memory_copy(vga_address, (vga_charset_saved + 3 * (VGA_CHARSET_SIZE / VGA_PLANES)), (VGA_CHARSET_SIZE / VGA_PLANES));
		}
	}



/****************************************************************************
 * vga_initialize ***********************************************************
 ****************************************************************************/
const sd	vga_initialize(
	tvidinit			*options)

	{
	tcmap				*vga_colormap = options->colormap;

	if (vga_state & VGA_INITIALIZED) return (-1);
	vga_state |= VGA_INITIALIZED;

	/* check cmdline given vga mode */

	if (vga_cmdline_mode_string != NULL)
		{
		d			mode;
		d			result;

		for (mode = 0, result = 1; vga_modes[mode].mode_string != NULL; mode++)
			{
			if ((result = string_compare(vga_cmdline_mode_string, vga_modes[mode].mode_string)) == 0) break;
			}
		if (result != 0) return (-1);
		vga_mode = mode;
		}

	/* get access to the vga i/o ports */

	if (ec64sys_ioperm(VGA_IO_BASE, VGA_IO_SIZE, 1) == -1) return (-1);
	vga_state |= VGA_IOPERM;

	/* check if we have color or monochrome register area */

	if (!(ec64sys_inb(VGA_MISC_DATA_READ) & 0x01))
		{
		if (ec64sys_ioperm(VGA_IO_BASE_MONOCHROME1, VGA_IO_SIZE_MONOCHROME1, 1) == -1) return (-1);
		vga_state |= VGA_IOPERM_MONOCHROME1;
		if (ec64sys_ioperm(VGA_IO_BASE_MONOCHROME2, VGA_IO_SIZE_MONOCHROME2, 1) == -1) return (-1);
		vga_state |= VGA_IOPERM_MONOCHROME2;
		vga_crt_index = VGA_CRT_INDEX_MONOCHROME;
		vga_crt_data = VGA_CRT_DATA_MONOCHROME;
		vga_input_status = VGA_INPUT_STATUS_MONOCHROME;
		}

	/* save the current settings */

	vga_save_registers();

	/* mmap the video memory */

		{
		tmmap			vga_mmap =
					{
					0,
					0,
					(PROT_READ | PROT_WRITE),
					MAP_SHARED,
					0,
					0xa0000
					};

		vga_mmap.length = vga_mmap_size = 0x10000;
		if ((vga_mmap.handle = ec64sys_open(STRING_TYPE "/dev/mem", O_RDWR)) == -1)
			{
			vga_restore();
			error_text_message_exit(STRING_TYPE "couldn't open /dev/mem");
			}
		if ((vga_address = (b *) ec64sys_mmap(&vga_mmap)) == NULL)
			{
			vga_restore();
			error_text_exit(STRING_TYPE "mmap() failed");
			}
		vga_state |= VGA_MMAP;
		ec64sys_close(vga_mmap.handle);
		}

	/* initialize the vga with our values */

	vga_switch_mode();
	vga_set_colormap(vga_colormap);
	vga_x_resolution = vga_modes[vga_mode].x_resolution;
	vga_y_resolution = vga_modes[vga_mode].y_resolution;
	vga_virtual_x_resolution = vga_modes[vga_mode].x_resolution;
	vga_c64_x_offset = 0;
	vga_c64_y_offset = 0;

	/* save the current content of the videomemory (this is done to
	   preserve the textmode charset, though the charset itself only
	   occupies a part of this memory, but i don't know the correct
	   location so i save the whole 256 KB) */

	vga_save_charset();

	/* set the options */

	options->x_resolution = options->c64_x_resolution = vga_x_resolution;
	options->y_resolution = options->c64_y_resolution = vga_y_resolution;
	options->virtual_x_resolution = vga_virtual_x_resolution;
	options->c64_x_offset = vga_c64_x_offset = 0;
	options->c64_y_offset = vga_c64_y_offset = 0;
	options->c64_screen_buffer = vga_screen_buffer;
	options->c64_screen_lines = vga_screen_lines;

	/* exit */

	return(0);
	}



/****************************************************************************
 * vga_restore **************************************************************
 ****************************************************************************/
void	vga_restore(
	void)

	{
	vga_restore_charset();
	vga_restore_registers();
	if (vga_state & VGA_MMAP)
		{
		ec64sys_munmap(vga_address, vga_mmap_size);
		vga_state &= ~ VGA_MMAP;
		}
	if (vga_state & VGA_IOPERM_MONOCHROME2)
		{
		ec64sys_ioperm(VGA_IO_BASE_MONOCHROME2, VGA_IO_SIZE_MONOCHROME2, 0);
		vga_state &= ~ VGA_IOPERM_MONOCHROME2;
		}
	if (vga_state & VGA_IOPERM_MONOCHROME1)
		{
		ec64sys_ioperm(VGA_IO_BASE_MONOCHROME1, VGA_IO_SIZE_MONOCHROME1, 0);
		vga_state &= ~ VGA_IOPERM_MONOCHROME1;
		}
	if (vga_state & VGA_IOPERM)
		{
		ec64sys_ioperm(VGA_IO_BASE, VGA_IO_SIZE, 0);
		vga_state &= ~ VGA_IOPERM;
		}
	vga_state &= ~ VGA_INITIALIZED;
	}



/****************************************************************************
 * vga_set_mode *************************************************************
 ****************************************************************************/
void	vga_set_mode_string(
	const b				*vga_mode_string)

	{
	vga_cmdline_mode_string = vga_mode_string;
	}



/****************************************************************************
 * vga_set_options **********************************************************
 ****************************************************************************/
void	vga_set_options(
	tvidinit			*options)

	{
	vga_c64_x_offset = options->c64_x_offset;
	vga_c64_y_offset = options->c64_y_offset;
	options->virtual_x_resolution = vga_virtual_x_resolution;
	options->c64_screen_buffer = vga_screen_buffer;
	options->c64_screen_lines = vga_screen_lines;
	}



/****************************************************************************
 * vga_draw_text ************************************************************
 ****************************************************************************/
void	vga_draw_text(
	const d				x_start,
	const d				y_start,
	const b				*text)

	{
	b				*buffer = vga_address + ((y_start * vga_virtual_x_resolution) / VGA_PLANES) + (x_start / VGA_PLANES);
	d				x_length = (DRAW_X_CHARACTER_SIZE * string_length(text)) / VGA_PLANES;
	d				plane;
	d				line;

	draw_text(vga_draw_buffer, vga_virtual_x_resolution, text);
	for (plane = 0; plane < VGA_PLANES; plane ++)
		{
		vga_select_write_plane((1 << plane), 0);
		for (line = 0; line < DRAW_Y_CHARACTER_SIZE; line += 2)
			draw_planes(
				(buffer + (vga_virtual_x_resolution / VGA_PLANES) * line),
				(vga_draw_buffer + plane + vga_virtual_x_resolution * line),
				(vga_virtual_x_resolution / 4),
				vga_virtual_x_resolution,
				x_length);
		}
	}



/****************************************************************************
 * vga_draw_led *************************************************************
 ****************************************************************************/
void	vga_draw_led(
	const d				x_start,
	const d				y_start,
	const b				led_state)

	{
	b				*buffer = vga_address + ((y_start * vga_virtual_x_resolution) / VGA_PLANES) + (x_start / VGA_PLANES);
	d				plane;
	d				line;

	draw_led(vga_draw_buffer, vga_virtual_x_resolution, led_state);
	for (plane = 0; plane < VGA_PLANES; plane ++)
		{
		vga_select_write_plane((1 << plane), 0);
		for (line = 0; line < DRAW_Y_LED_SIZE; line += 2)
			draw_planes(
				(buffer + (vga_virtual_x_resolution / VGA_PLANES) * line),
				(vga_draw_buffer + plane + vga_virtual_x_resolution * line),
				(vga_virtual_x_resolution / VGA_PLANES),
				vga_virtual_x_resolution,
				(DRAW_X_LED_SIZE / VGA_PLANES));
		}
	}



/****************************************************************************
 * vga_draw_c64_screen ******************************************************
 ****************************************************************************/
void	vga_draw_c64_screen(
	void)

	{
	b				*buffer = vga_address + ((vga_c64_y_offset * vga_virtual_x_resolution) / VGA_PLANES) + (vga_c64_x_offset / VGA_PLANES);
	d				line1;
	d				line2;
	d				plane;
	d				offset = 0;

	/*
	!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
	TODO: only draw the real c64_x_resolution, at the moment i draw
	      vga_virtual_x_resolution - vga_c64_x_offset, this is ok if
	      c64_x_resolution is nearly equal to vga_x_resolution
	!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
	*/

	while ((line1 = vga_screen_lines[offset]) != 0)
		{
		offset ++;
		line1 --;
		if ((line2 = vga_screen_lines[offset]) != 0)
			{
			offset ++;
			line2 --;
			if (line1 >= line2)
				error_text_exit(STRING_TYPE "!INTERNAL FAILURE! vga.c: got wrong line format from core");
			for (plane = 0; plane < VGA_PLANES; plane ++)
				{
				vga_select_write_plane((1 << plane), 0);
				draw_planes(
					(buffer + (vga_virtual_x_resolution / VGA_PLANES) * line1),
					(vga_screen_buffer + plane + vga_virtual_x_resolution * line1),
					((vga_virtual_x_resolution / VGA_PLANES) * (line2 - line1)),
					(vga_virtual_x_resolution * (line2 - line1)),
					((vga_virtual_x_resolution - vga_c64_x_offset) / VGA_PLANES));
				}
			}
		else
			{
			for (plane = 0; plane < VGA_PLANES; plane ++)
				{
				vga_select_write_plane((1 << plane), 0);
				draw_plane(
					(buffer + (vga_virtual_x_resolution / VGA_PLANES) * line1),
					(vga_screen_buffer + plane + vga_virtual_x_resolution * line1),
					((vga_virtual_x_resolution - vga_c64_x_offset) / VGA_PLANES));
				}
			}
		}
	}



/****************************************************************************
 * vga_area_clear ***********************************************************
 ****************************************************************************/
void	vga_area_clear(
	const d				x_start,
	const d				y_start,
	const d				x_end,
	const d				y_end)

	{
	if (x_start < x_end)
		{
		b			*buffer = vga_address + ((y_start * vga_virtual_x_resolution) / VGA_PLANES) + (x_start / VGA_PLANES);
		d			x_length = (x_end - x_start) / VGA_PLANES;
		d			line;

		vga_select_write_plane(0x0f, 0);
		for (line = y_start; line < y_end; line ++)
			{
			memory_fill(
				buffer,
				x_length,
				0);
			buffer += vga_virtual_x_resolution / VGA_PLANES;
			}
		}
	}



/****************************************************************************
 * vga_area_move ************************************************************
 ****************************************************************************/
void	vga_area_move(
	const d				x_start,
	const d				y_start,
	const d				x_end,
	const d				y_end,
	const d				x_destination,
	const d				y_destination)

	{
	if (x_start != x_destination)
		error_text_exit(STRING_TYPE "!INTERNAL FAILURE! vga.c: vga_area_move() failed");
	if (x_start < x_end)
		{
		d			line;
		d			x_length = (x_end - x_start) / VGA_PLANES;

		vga_select_write_plane(0x0f, 1);
		if (y_start < y_destination)
			{
			b		*source_buffer = vga_address + (((y_end - 1) * vga_virtual_x_resolution) / VGA_PLANES) + (x_start / VGA_PLANES);
			b		*destination_buffer = vga_address + (((y_destination + y_end - y_start - 1) * vga_virtual_x_resolution) / VGA_PLANES) + (x_destination / VGA_PLANES);

			for (line = y_end - 1; line >= y_start; line --)
				{
				memory_copy(destination_buffer, source_buffer, x_length);
				source_buffer -= vga_virtual_x_resolution / VGA_PLANES;
				destination_buffer -= vga_virtual_x_resolution / VGA_PLANES;
				}
			}
		else
			{
			b		*source_buffer = vga_address + ((y_start * vga_virtual_x_resolution) / VGA_PLANES) + (x_start / VGA_PLANES);
			b		*destination_buffer = vga_address + ((y_destination * vga_virtual_x_resolution) / VGA_PLANES)+ (x_destination / VGA_PLANES);

			for (line = y_start; line < y_end; line ++)
				{
				memory_copy(destination_buffer, source_buffer, x_length);
				source_buffer += vga_virtual_x_resolution / VGA_PLANES;
				destination_buffer += vga_virtual_x_resolution / VGA_PLANES;
				}
			}
		}
	}



#ifdef VGA_ENABLE_SCREENSHOT_SUPPORT
/****************************************************************************
 * vga_get_line *************************************************************
 ****************************************************************************/
void	vga_get_line(
	b				*buffer,
	const d				line,
	const d				length)

	{
	d				plane;
	d				counter;

	if (line > vga_y_resolution) return;
	for (plane = 0; plane < VGA_PLANES; plane ++)
		{
		vga_select_read_plane(plane, 0);
		for (counter = 0; counter < (length / VGA_PLANES); counter ++)
			buffer[VGA_PLANES * counter + plane] = vga_address[counter + (vga_virtual_x_resolution / VGA_PLANES) * line];
		}
	}
#endif /* VGA_ENABLE_SCREENSHOT_SUPPORT */
#endif /* VGA_ENABLE */
/******************************************************** ec64@unusedino.de */

